"""
tn3270.emulator
~~~~~~~~~~~~~~~
"""

from enum import Enum
from itertools import chain
import struct
import logging

from .datastream import Command, WCC, Order, AID, parse_outbound_message, \
                        format_inbound_read_buffer_message, \
                        format_inbound_read_modified_message, \
                        parse_orders, format_inbound_structured_fields
from .attributes import Attribute, AllExtendedAttribute, \
                        Highlight, HighlightExtendedAttribute, \
                        Color, ForegroundColorExtendedAttribute
from .structured_fields import StructuredField, ReadPartitionType, \
                               QueryListRequestType, QueryCode
from .ebcdic import DUP, FM

class CharacterSet(Enum):
    """Display cell character set."""

    GE = 1

class CellFormatting:
    """Display cell formatting."""

    def __init__(self, formatting=None, extended_attributes=None):
        self.blink = False
        self.reverse = False
        self.underscore = False
        self.color = 0x00

        if formatting is not None:
            self.blink = formatting.blink
            self.reverse = formatting.reverse
            self.underscore = formatting.underscore
            self.color = formatting.color

        if extended_attributes is not None:
            for extended_attribute in extended_attributes:
                self._apply_extended_attribute(extended_attribute)

    def _apply_extended_attribute(self, extended_attribute):
        if isinstance(extended_attribute, AllExtendedAttribute):
            self.blink = False
            self.reverse = False
            self.underscore = False
            self.color = 0x00
        elif isinstance(extended_attribute, HighlightExtendedAttribute):
            self.blink = extended_attribute.blink
            self.reverse = extended_attribute.reverse
            self.underscore = extended_attribute.underscore
        elif isinstance(extended_attribute, ForegroundColorExtendedAttribute):
            self.color = extended_attribute.color

    def __eq__(self, other):
        if not isinstance(other, CellFormatting):
            return False

        return other.blink == self.blink and other.reverse == self.reverse and \
               other.underscore == self.underscore and other.color == self.color

class Cell:
    """A display cell."""

    def __init__(self, formatting):
        self.formatting = formatting

class AttributeCell(Cell):
    """A attribute display cell."""

    def __init__(self, attribute, formatting=None):
        super().__init__(formatting)

        self.attribute = attribute

class CharacterCell(Cell):
    """A character display cell."""

    def __init__(self, byte, character_set=None, formatting=None):
        super().__init__(formatting)

        self.byte = byte
        self.character_set = character_set

class OperatorError(Exception):
    """Operator error."""

class ProtectedCellOperatorError(OperatorError):
    """Protected cell error."""

class FieldOverflowOperatorError(OperatorError):
    """Field overflow error."""

class Emulator:
    """TN3270 emulator."""

    def __init__(self, stream, rows, columns, supported_colors=8,
            supported_highlights=[Highlight.BLINK, Highlight.REVERSE,
                Highlight.UNDERSCORE]):
        self.logger = logging.getLogger(__name__)

        # TODO: Validate that stream has read_multiple() and write() methods.
        self.stream = stream

        if rows < 24:
            raise ValueError('Invalid rows, must be at least 24')

        if columns < 80:
            raise ValueError('Invalid columns, must be at least 80')

        if supported_colors not in [1, 4, 8]:
            raise ValueError('Invalid suported colors, must be 1, 4 or 8')

        self.default_dimensions = (24, 80)
        self.alternate_dimensions = (rows, columns)

        self.supported_colors = supported_colors
        self.supported_highlights = supported_highlights

        (self.rows, self.columns) = self.default_dimensions
        self.alternate = False

        self.cells = [CharacterCell(0x00) for index in range(rows * columns)]
        self.dirty = set(range(rows * columns))

        self.address = 0
        self.cursor_address = 0

        self.current_aid = AID.NONE
        self.keyboard_locked = True

    def update(self, **kwargs):
        """Read and execute outbound messages."""
        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug('Update')

        records = self.stream.read_multiple(**kwargs)

        if not records:
            return False

        for bytes_ in records:
            if self.logger.isEnabledFor(logging.DEBUG):
                self.logger.debug(f'\tRecord = {bytes_}')

            (command, *options) = parse_outbound_message(bytes_)

            self._execute(command, *options)

        return True

    def aid(self, aid):
        """AID key."""
        if aid == AID.CLEAR:
            self._clear()

        self.current_aid = aid
        self.keyboard_locked = True

        self._read_modified()

    def tab(self, direction=1):
        """Tab or backtab key."""
        address = self._calculate_tab_address(self.cursor_address, direction)

        if address is not None:
            self.cursor_address = address

    def newline(self):
        """Move to the next line or the subsequent unprotected field."""
        current_row = self.cursor_address // self.columns

        address = self._wrap_address((current_row + 1) * self.columns)

        (attribute, attribute_address) = self.find_attribute(address)

        if attribute is not None and not attribute.protected and attribute_address != address:
            self.cursor_address = address
            return

        address = self._calculate_tab_address(address, direction=1)

        if address is not None:
            self.cursor_address = address

    def home(self):
        """Home key."""
        if not self.is_formatted():
            self.cursor_address = 0
            return

        addresses = self._get_addresses(0, (self.rows * self.columns) - 1)

        address = next((address for address in addresses
                        if isinstance(self.cells[address], AttributeCell)
                        and not self.cells[address].attribute.protected), None)

        if address is not None:
            self.cursor_address = self._wrap_address(address + 1)

    def cursor_up(self):
        """Cursor up key."""
        self.cursor_address = self._wrap_address(self.cursor_address - self.columns)

    def cursor_down(self):
        """Cursor down key."""
        self.cursor_address = self._wrap_address(self.cursor_address + self.columns)

    def cursor_left(self, rate=1):
        """Cursor left key."""
        if rate < 1 or rate > 2:
            raise ValueError('Invalid rate')

        self.cursor_address = self._wrap_address(self.cursor_address - rate)

    def cursor_right(self, rate=1):
        """Cursor right key."""
        if rate < 1 or rate > 2:
            raise ValueError('Invalid rate')

        self.cursor_address = self._wrap_address(self.cursor_address + rate)

    def input(self, byte, insert=False):
        """Single character input."""
        self._input(byte, insert=insert)

    def dup(self, insert=False):
        """Duplicate (DUP) key."""
        self._input(DUP, insert=insert, move=False)

        # TODO: Moving to the next unprotected field should be reusable - should the
        # calculate_tab_address method be refactored to be more generic or at least
        # a single next_unprotected filter?
        addresses = self._get_addresses(self.cursor_address,
                                        self._wrap_address(self.cursor_address - 1))

        address = next((address for address in addresses
                        if isinstance(self.cells[address], AttributeCell)
                        and not self.cells[address].attribute.protected), None)

        if address is not None:
            self.cursor_address = self._wrap_address(address + 1)

    def field_mark(self, insert=False):
        """Field mark (FM) key."""
        self._input(FM, insert=insert)

    def backspace(self):
        """Backspace key."""
        if isinstance(self.cells[self.cursor_address], AttributeCell):
            raise ProtectedCellOperatorError

        (start_address, end_address, attribute) = self.get_field(self.cursor_address)

        if self.cursor_address == start_address:
            return

        self._shift_left(self._wrap_address(self.cursor_address - 1), end_address)

        if attribute is not None:
            attribute.modified = True

        self.cursor_address = self._wrap_address(self.cursor_address - 1)

    def delete(self):
        """Delete key."""
        if not self.is_formatted():
            return

        if isinstance(self.cells[self.cursor_address], AttributeCell):
            raise ProtectedCellOperatorError

        (_, end_address, attribute) = self.get_field(self.cursor_address)

        self._shift_left(self.cursor_address, end_address)

        if attribute is not None:
            attribute.modified = True

    def erase_end_of_field(self):
        """Erase end of field (EOF) key."""
        if self.is_formatted():
            if isinstance(self.cells[self.cursor_address], AttributeCell):
                raise ProtectedCellOperatorError

            (_, end_address, attribute) = self.get_field(self.cursor_address)

            for address in self._get_addresses(self.cursor_address, end_address):
                self._write_character(address, 0x00, preserve=True)

            attribute.modified = True
        else:
            for address in self._get_addresses(self.cursor_address, (self.rows * self.columns) - 1):
                self._write_character(address, 0x00, preserve=True)

    def erase_input(self):
        """Erase input key."""
        if self.is_formatted():
            for (start_address, end_address, attribute) in self.get_fields():
                for address in self._get_addresses(start_address, end_address):
                    self._write_character(address, 0x00, preserve=True)

                attribute.modified = False

            self.cursor_address = self._calculate_tab_address(0, 1) or 0
        else:
            self._clear()

    def get_bytes(self, start_address, end_address):
        """Get character cell bytes."""
        addresses = self._get_addresses(start_address, end_address)

        return bytes([self.cells[address].byte if isinstance(self.cells[address], CharacterCell) else 0x00 for address in addresses])

    def is_formatted(self):
        """Is there at least one attribute?"""
        return any([isinstance(cell, AttributeCell) for cell in self.cells])

    def get_field(self, address):
        """Get the unprotected field containing or starting at the address."""
        (attribute, start_attribute_address) = self.find_attribute(address)

        if attribute is None or attribute.protected:
            raise ProtectedCellOperatorError

        start_address = self._wrap_address(start_attribute_address + 1)

        # By using the field start attribute address as the search end address we know
        # there will be at least one attribute byte found even in the case of a single
        # field.
        addresses = self._get_addresses(start_address, start_attribute_address)

        end_attribute_address = next((address for address in addresses
                                      if isinstance(self.cells[address], AttributeCell)))

        end_address = self._wrap_address(end_attribute_address - 1)

        return (start_address, end_address, attribute)

    def get_fields(self):
        """Get all unprotected fields."""
        fields = []

        for address in range(0, self.rows * self.columns):
            cell = self.cells[address]

            if isinstance(cell, AttributeCell) and not cell.attribute.protected:
                field = self.get_field(address)

                fields.append(field)

        return fields

    def find_attribute(self, address):
        """Find the applicable attribute for the address."""
        addresses = self._get_addresses(address, self._wrap_address(address + 1),
                                        direction=-1)

        for address in addresses:
            cell = self.cells[address]

            if isinstance(cell, AttributeCell):
                return (cell.attribute, address)

        return (None, None)

    def alarm(self):
        """Alarm stub."""
        pass

    def _execute(self, command, *options):
        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug('Execute')
            self.logger.debug(f'\tCommand = {command}')

        if command == Command.W:
            self._write(*options)
        elif command == Command.RB:
            self._read_buffer()
        elif command == Command.NOP:
            pass
        elif command in [Command.EW, Command.EWA]:
            self._erase(command == Command.EWA)
            self._write(*options)
        elif command == Command.RM:
            self._read_modified()
        elif command == Command.RMA:
            self._read_modified(all_=True)
        elif command == Command.EAU:
            self._erase_all_unprotected()
        elif command == Command.WSF:
            self._write_structured_fields(*options)

    def _erase(self, alternate):
        self.logger.debug('Erase')
        self.logger.debug(f'\tAlternate = {alternate}')

        self._clear()

        if alternate:
            (self.rows, self.columns) = self.alternate_dimensions
        else:
            (self.rows, self.columns) = self.default_dimensions

        self.alternate = alternate

    def _erase_all_unprotected(self):
        self.logger.debug('Erase All Unprotected')

        if self.is_formatted():
            for (start_address, end_address, attribute) in self.get_fields():
                for address in self._get_addresses(start_address, end_address):
                    self._write_character(address, 0x00, preserve=True)

                attribute.modified = False

            self.cursor_address = self._calculate_tab_address(0, 1) or 0
        else:
            self._clear()

        self.current_aid = AID.NONE
        self.keyboard_locked = False

    def _write(self, wcc, orders):
        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug('Write')
            self.logger.debug(f'\tWCC = {wcc}')

        if wcc.reset_modified:
            for cell in self.cells:
                if isinstance(cell, AttributeCell):
                    cell.attribute.modified = False

        field_formatting = None
        character_formatting = None
        character_set = None

        for (order, data) in orders:
            if self.logger.isEnabledFor(logging.DEBUG):
                if order is None:
                    self.logger.debug(f'\t{data}')
                else:
                    self.logger.debug(f'\t{order}')
                    self.logger.debug(f'\t\tParameters = {data}')

            if order == Order.PT:
                # TODO: Implement additional PT cases
                if isinstance(self.cells[self.address], AttributeCell) and not self.cells[self.address].attribute.protected:
                    self.address = self._wrap_address(self.address + 1)
                else:
                    raise NotImplementedError('PT order is not fully supported')
            elif order == Order.GE:
                self._write_character(self.address, data[0], CharacterSet.GE, None)

                self.address = self._wrap_address(self.address + 1)
            elif order == Order.SBA:
                if data[0] >= (self.rows * self.columns):
                    self.logger.warning(f'Address {data[0]} is out of range')
                else:
                    self.address = data[0]
            elif order == Order.EUA:
                stop_address = data[0]

                # TODO: Validate stop_address is in range...
                end_address = self._wrap_address(stop_address - 1)

                addresses = self._get_addresses(self.address, end_address)
                unprotected_addresses = self._get_unprotected_addresses()

                for address in unprotected_addresses.intersection(addresses):
                    self._write_character(address, 0x00, preserve=True)

                self.address = stop_address
            elif order == Order.IC:
                self.cursor_address = self.address
            elif order == Order.SF:
                field_formatting = None

                self._write_attribute(self.address, data[0], field_formatting)

                self.address = self._wrap_address(self.address + 1)
            elif order == Order.SA:
                character_formatting = CellFormatting(character_formatting, extended_attributes=[data[0]])
            elif order == Order.SFE:
                # TODO: Confirm that formatting should be reset here
                field_formatting = CellFormatting(None, extended_attributes=data[1])

                self._write_attribute(self.address, data[0] or Attribute(0), field_formatting)

                self.address = self._wrap_address(self.address + 1)
            elif order == Order.MF:
                if isinstance(self.cells[self.address], AttributeCell):
                    (attribute, extended_attributes) = data

                    if attribute is not None:
                        self.cells[self.address].attribute = attribute

                    if extended_attributes:
                        existing_formatting = self.cells[self.address].formatting

                        # TODO: Confirm that this should affect "global" formatting...
                        field_formatting = CellFormatting(existing_formatting, extended_attributes)

                        self.cells[self.address].formatting = field_formatting

                    self.address = self._wrap_address(self.address + 1)
                else:
                    self.logger.warning('MF order rejected as cell is not attribute cell')
            elif order == Order.RA:
                (stop_address, byte, is_ge) = data

                # TODO: Validate stop_address is in range...
                end_address = self._wrap_address(stop_address - 1)

                addresses = self._get_addresses(self.address, end_address)

                for address in addresses:
                    self._write_character(address, byte, CharacterSet.GE if is_ge else character_set, character_formatting)

                self.address = stop_address
            elif order is None:
                for byte in data:
                    self._write_character(self.address, byte, character_set, character_formatting)

                    self.address = self._wrap_address(self.address + 1)

        if wcc.unlock_keyboard:
            self.current_aid = AID.NONE
            self.keyboard_locked = False

        if wcc.alarm:
            self.alarm()

    def _clear(self):
        for address in range(self.rows * self.columns):
            self._write_character(address, 0x00, None, None)

        self.address = 0
        self.cursor_address = 0

    def _read_buffer(self):
        orders = self._generate_inbound_orders(0, (self.rows * self.columns) - 1)

        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug('Read Buffer')
            self.logger.debug(f'\tAID    = {self.current_aid}')
            self.logger.debug(f'\tOrders = {orders}')

        bytes_ = format_inbound_read_buffer_message(self.current_aid, self.cursor_address, orders)

        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug(f'\tData   = {bytes_}')

        self.stream.write(bytes_)

    def _read_modified(self, all_=False):
        if self.is_formatted():
            modified_field_ranges = [(start_address, end_address) for (start_address, end_address, attribute) in self.get_fields() if attribute.modified]
        else:
            modified_field_ranges = [(0, (self.rows * self.columns) - 1)]

        orders = []

        for (start_address, end_address) in modified_field_ranges:
            orders.append((Order.SBA, [start_address]))

            orders.extend(self._generate_inbound_orders(start_address, end_address, filter_null=True))

        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug('Read Modified')
            self.logger.debug(f'\tAID    = {self.current_aid}')
            self.logger.debug(f'\tOrders = {orders}')
            self.logger.debug(f'\tAll    = {all_}')

        bytes_ = format_inbound_read_modified_message(self.current_aid, self.cursor_address, orders, all_)

        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug(f'\tData   = {bytes_}')

        self.stream.write(bytes_)

    def _input(self, byte, insert=False, move=True):
        attribute = None

        if self.is_formatted():
            if isinstance(self.cells[self.cursor_address], AttributeCell):
                raise ProtectedCellOperatorError

            (_, end_address, attribute) = self.get_field(self.cursor_address)

            # TODO: Implement numeric field validation.

            if insert and self.cells[self.cursor_address].byte != 0x00:
                addresses = self._get_addresses(self.cursor_address, end_address)

                first_null_address = next((address for address in addresses
                                           if self.cells[address].byte == 0x00), None)

                if first_null_address is None:
                    raise FieldOverflowOperatorError

                self._shift_right(self.cursor_address, first_null_address)
        else:
            if insert:
                raise NotImplementedError('Insert input on unformatted screen is not supported')

        self._write_character(self.cursor_address, byte, preserve=True)

        if attribute is not None:
            attribute.modified = True

        if not move:
            return

        self.cursor_address = self._wrap_address(self.cursor_address + 1)

        # TODO: Is this correct - does this only happen if skip?
        if isinstance(self.cells[self.cursor_address], AttributeCell):
            skip = self.cells[self.cursor_address].attribute.skip

            addresses = self._get_addresses(self.cursor_address,
                                            self._wrap_address(self.cursor_address - 1))

            address = next((address for address in addresses
                            if isinstance(self.cells[address], AttributeCell)
                            and (not skip or (skip and not self.cells[address].attribute.protected))), None)

            if address is not None:
                self.cursor_address = self._wrap_address(address + 1)

    def _get_addresses(self, start_address, end_address, direction=1):
        if direction < 0:
            if end_address > start_address:
                return chain(reversed(range(0, start_address + 1)),
                             reversed(range(end_address, self.rows * self.columns)))

            return reversed(range(end_address, start_address + 1))

        if end_address < start_address:
            return chain(range(start_address, self.rows * self.columns),
                         range(0, end_address + 1))

        return range(start_address, end_address + 1)

    def _wrap_address(self, address):
        if address < 0 or address >= (self.rows * self.columns):
            return address % (self.rows * self.columns)

        return address

    def _get_unprotected_addresses(self):
        addresses = set()

        for (start_address, end_address, _) in self.get_fields():
            addresses.update(self._get_addresses(start_address, end_address))

        return addresses

    def _calculate_tab_address(self, address, direction):
        if direction < 0:
            if address > 0 and isinstance(self.cells[address - 1], AttributeCell):
                address -= 1

            start_address = self._wrap_address(address - 1)
            end_address = self._wrap_address(address)
        else:
            start_address = self._wrap_address(address)
            end_address = self._wrap_address(address - 1)

        addresses = self._get_addresses(start_address, end_address, direction)

        address = next((address for address in addresses
                        if isinstance(self.cells[address], AttributeCell)
                        and not self.cells[address].attribute.protected), None)

        if address is None:
            return None

        return self._wrap_address(address + 1)

    def _write_attribute(self, index, attribute, formatting=None, preserve=False):
        cell = self.cells[index]

        if preserve:
            formatting = cell.formatting

        if isinstance(cell, AttributeCell):
            if cell.attribute.value == attribute.value and cell.formatting == formatting:
                return False

            cell.attribute = attribute
            cell.formatting = formatting
        else:
            self.cells[index] = AttributeCell(attribute, formatting)

        self.dirty.add(index)

        return True

    def _write_character(self, index, byte, character_set=None, formatting=None, preserve=False):
        cell = self.cells[index]

        if preserve:
            character_set = cell.character_set
            formatting = cell.formatting

        if isinstance(cell, CharacterCell):
            if cell.byte == byte and cell.character_set == character_set and cell.formatting == formatting:
                return False

            cell.byte = byte
            cell.character_set = character_set
            cell.formatting = formatting
        else:
            self.cells[index] = CharacterCell(byte, character_set, formatting)

        self.dirty.add(index)

        return True

    def _shift_left(self, start_address, end_address):
        addresses = list(self._get_addresses(start_address, end_address))

        for (left_address, right_address) in zip(addresses, addresses[1:]):
            self._write_character(left_address, self.cells[right_address].byte, preserve=True)

        self._write_character(end_address, 0x00, preserve=True)

    def _shift_right(self, start_address, end_address):
        addresses = list(self._get_addresses(start_address, end_address))

        for (left_address, right_address) in reversed(list(zip(addresses, addresses[1:]))):
            self._write_character(right_address, self.cells[left_address].byte, preserve=True)

        self._write_character(start_address, 0x00, preserve=True)

    def _generate_inbound_orders(self, start_address, end_address, filter_null=False):
        orders = []

        data = bytearray()

        def eject():
            nonlocal data

            if data:
                orders.append((None, data))

                data = bytearray()

        for address in self._get_addresses(start_address, end_address):
            cell = self.cells[address]

            if isinstance(cell, AttributeCell):
                eject()

                orders.append((Order.SF, [cell.attribute]))
            elif isinstance(cell, CharacterCell):
                if not filter_null or cell.byte != 0x00:
                    if cell.character_set == CharacterSet.GE:
                        eject()

                        orders.append((Order.GE, [cell.byte]))
                    else:
                        data.append(cell.byte)

        eject()

        return orders

    def _write_structured_fields(self, structured_fields):
        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug('Write Structured Fields')
            self.logger.debug(f'\tFields = {structured_fields}')

        for (id_, data) in structured_fields:
            if id_ == StructuredField.READ_PARTITION:
                self._read_partition(data)
            elif id_ == StructuredField.ERASE_RESET:
                self._erase(data[0] == 0x80)
            elif id_ == StructuredField.OUTBOUND_3270DS:
                self._outbound_3270ds(data)
            else:
                raise NotImplementedError(f'Structured field 0x{id_:02x} not supported')

    def _read_partition(self, data):
        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug('Read Partition (Structured Field)')
            self.logger.debug(f'\tData = {data}')

        partition = data[0]
        type_ = data[1]

        if type_ == ReadPartitionType.QUERY:
            if partition != 0xff:
                self.logger.warning(f'Partition should be 0xff for query, received 0x{partition:02x}')

            self._query()
        elif type_ == ReadPartitionType.QUERY_LIST:
            if partition != 0xff:
                self.logger.warning(f'Partition should be 0xff for query list, received 0x{partition:02x}')

            request_type = QueryListRequestType(data[2])

            request_codes = None

            if request_type != QueryListRequestType.ALL:
                request_codes = [QueryCode(code) for code in data[3:]]

            self._query(request_type, request_codes)
        else:
            raise NotImplementedError(f'Read partition type 0x{type_:02x} not supported')

    def _outbound_3270ds(self, data):
        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug('Outbound 3270 DS (Structured Field)')
            self.logger.debug(f'\tData = {data}')

        partition = data[0]
        command = data[1]

        if partition != 0x00:
            self.logger.warning(f'Partition 0x{partition:02x} not supported')

        if command == 0xf1:
            self._write(WCC(data[2]), parse_orders(data[3:]))
        elif command in [0xf5, 0x7e]:
            self._erase(command == 0x7e)
            self._write(WCC(data[2]), parse_orders(data[3:]))
        elif command == 0x6f:
            self._erase_all_unprotected()
        else:
            raise NotImplementedError(f'Outbound 3270 DS command 0x{command:02x} not supported')

    def _query(self, request_type=None, request_codes=None):
        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug('Query')
            self.logger.debug(f'\tRequest Type  = {request_type}')
            self.logger.debug(f'\tRequest Codes = {request_codes}')

        codes = [
            QueryCode.USABLE_AREA, QueryCode.ALPHANUMERIC_PARTITIONS,
            QueryCode.COLOR, QueryCode.HIGHLIGHT,
            QueryCode.REPLY_MODES, QueryCode.IMPLICIT_PARTITIONS
        ]

        if request_type == QueryListRequestType.LIST:
            codes = request_codes
        elif request_type == QueryListRequestType.EQUIVALENT_AND_LIST:
            if request_codes is not None:
                codes = set(codes + request_codes)

        # Generate the replies.
        replies = []

        for code in codes:
            reply = None

            if code == QueryCode.USABLE_AREA:
                reply = _query_usable_area(self.alternate_dimensions)
            elif code == QueryCode.ALPHANUMERIC_PARTITIONS:
                reply = _query_alphanumeric_partitions(self.alternate_dimensions)
            elif code == QueryCode.COLOR:
                reply = _query_color(self.supported_colors)
            elif code == QueryCode.HIGHLIGHT:
                reply = _query_highlight(self.supported_highlights)
            elif code == QueryCode.REPLY_MODES:
                reply = _query_reply_modes()
            elif code == QueryCode.IMPLICIT_PARTITIONS:
                reply = _query_implicit_partitions(self.alternate_dimensions)

            if reply is not None:
                replies.append((code, reply))

        if request_type == QueryListRequestType.LIST and not replies:
            replies.append((QueryCode.NULL, None))

        # Generate the summary reply.
        structured_fields = [(StructuredField.QUERY_REPLY, bytes([QueryCode.SUMMARY, QueryCode.SUMMARY] + [code for (code, _) in replies]))]

        # Append the query replies.
        for (code, data) in replies:
            structured_fields.append((StructuredField.QUERY_REPLY, bytes([code]) + (data if data is not None else bytes([]))))

        bytes_ = format_inbound_structured_fields(structured_fields)

        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug(f'\tData = {bytes_}')

        self.stream.write(bytes_)

def _query_usable_area(dimensions):
    (rows, columns) = dimensions

    return struct.pack('>BBHHBHHHHBBH',
        0x01,           # 12/14-bit addressing allowed
        0x00,           # Cell units, no special features
        columns,        # Width in cells
        rows,           # Width in rows
        0x01,           # Millimeters
        10,             # X resolution fraction numerator
        741,            # X resolution fraction denominator
        2,              # Y resolution fraction numerator
        111,            # Y resolution fraction denominator
        9,              # Width of cell in millimeters
        12,             # Height of cell in millimeters
        rows * columns  # Buffer size
    )

def _query_alphanumeric_partitions(dimensions):
    (rows, columns) = dimensions

    return struct.pack('>BHB',
        1,              # One partition
        rows * columns, # Buffer size
        0x00            # No special features
    )

def _query_color(supported_colors):
    reply = struct.pack('>BBBB',
        0x00,           # Flags
        8,              # Colors
        0x00,           # Default
        Color.GREEN
    )

    for attribute in range(0xf1, 0xf7 + 1):
        reply += struct.pack('>BB', attribute, attribute if supported_colors > 4 else 0x00)

    return reply

def _query_highlight(supported_highlights):
    count = len(supported_highlights) + 1

    reply = struct.pack('>BBB',
        count,          # Types
        0x00,           # Default
        Highlight.NORMAL
    )

    for attribute in supported_highlights:
        reply += struct.pack('>BB', attribute, attribute)

    return reply

def _query_reply_modes():
    # TODO: Extended field mode is not really supported, but this is enough to trick z/VM
    # into sending us pretty screens.
    return struct.pack('>BB',
        0x00,           # Field mode
        0x01            # Extended field mode
    )

def _query_implicit_partitions(dimensions):
    (rows, columns) = dimensions

    return struct.pack('>HBBBHHHH',
        0x0000,         # Flags
        0x0b,           # Length
        0x01,           # Size
        0x00,           # Flags
        80,             # Width
        24,             # Height
        columns,        # Width
        rows            # Height
    )



from .__about__ import __version__

from .telnet import Telnet

from .datastream import AID

from .attributes import (
    Highlight,
    Color
)

from .emulator import (
    Emulator,
    AttributeCell,
    CharacterCell,
    CharacterSet,
    OperatorError,
    ProtectedCellOperatorError,
    FieldOverflowOperatorError
)



"""
tn3270.structured_fields
~~~~~~~~~~~~~~~~~~~~~~~~
"""

from enum import Enum, IntEnum

class StructuredField(IntEnum):
    READ_PARTITION = 0x01
    ERASE_RESET = 0x03
    OUTBOUND_3270DS = 0x40
    QUERY_REPLY = 0x81

class ReadPartitionType(IntEnum):
    QUERY = 0x02
    QUERY_LIST = 0x03

class QueryListRequestType(IntEnum):
    LIST = 0x00
    EQUIVALENT_AND_LIST = 0x40
    ALL = 0x80

class QueryCode(IntEnum):
    SUMMARY = 0x80
    USABLE_AREA = 0x81
    ALPHANUMERIC_PARTITIONS = 0x84
    COLOR = 0x86
    HIGHLIGHT = 0x87
    REPLY_MODES = 0x88
    IMPLICIT_PARTITIONS = 0xa6
    NULL = 0xff



"""
tn3270.attributes
~~~~~~~~~~~~~~~~~
"""

from enum import IntEnum

class Attribute:
    """Attribute."""

    def __init__(self, value):
        # TODO: Validate input - looks like there is a parity bit.
        self._original_value = value

        self.protected = bool(value & 0x20)
        self.numeric = bool(value & 0x10)
        self.skip = self.protected and self.numeric

        display = (value & 0x0c) >> 2

        self.intensified = (display == 2)
        self.hidden = (display == 3)

        self.modified = bool(value & 0x01)

    @property
    def value(self):
        # TODO: Reconstruct the entire attribute from parts, this assumes
        # modified is the only part that can change (which is true).
        return (self._original_value & 0xfe) | int(self.modified)

    def __repr__(self):
        return (f'<Attribute protected={self.protected}, numeric={self.numeric}, '
                f'skip={self.skip}, intensified={self.intensified}, '
                f'hidden={self.hidden}, modified={self.modified}>')

class ExtendedAttributeType(IntEnum):
    """Extended attribute type."""

    ALL = 0x00
    HIGHLIGHT = 0x41
    FOREGROUND_COLOR = 0x42

class ExtendedAttribute:
    """Extended attribute."""

    def __init__(self, type_, value):
        self.type_ = type_
        self.value = value

    def __repr__(self):
        return f'<ExtendedAttribute type={self.type_}, value={self.value}>'

class AllExtendedAttribute(ExtendedAttribute):
    """Reset all character attributes extended attribute."""

    def __init__(self, type_, value):
        super().__init__(type_, value)

class Highlight(IntEnum):
    """Highlight."""

    NORMAL = 0xf0
    BLINK = 0xf1
    REVERSE = 0xf2
    UNDERSCORE = 0xf4
    INTENSIFY = 0xf8

class HighlightExtendedAttribute(ExtendedAttribute):
    """Highlight extended attribute."""

    def __init__(self, type_, value):
        super().__init__(type_, value)

        self.blink = False
        self.reverse = False
        self.underscore = False

        if value == Highlight.BLINK:
            self.blink = True
        elif value == Highlight.REVERSE:
            self.reverse = True
        elif value == Highlight.UNDERSCORE:
            self.underscore = True

    def __repr__(self):
        return (f'<HighlightExtendedAttribute blink={self.blink}, '
                f'reverse={self.reverse}, underscore={self.underscore}>')

class Color(IntEnum):
    """Color."""

    NEUTRAL = 0x00
    BLUE = 0xf1
    RED = 0xf2
    PINK = 0xf3
    GREEN = 0xf4
    TURQUOISE = 0xf5
    YELLOW = 0xf6
    WHITE = 0xf7

class ForegroundColorExtendedAttribute(ExtendedAttribute):
    """Foreground extended attribute."""

    def __init__(self, type_, value):
        super().__init__(type_, value)

        self.color = value

    def __repr__(self):
        return f'<ForegroundColorExtendedAttribute color={self.color}>'



"""
tn3270.datastream
~~~~~~~~~~~~~~~~~
"""

from enum import Enum
import logging

from .attributes import Attribute, ExtendedAttributeType, ExtendedAttribute, \
                        AllExtendedAttribute, HighlightExtendedAttribute, \
                        ForegroundColorExtendedAttribute

logger = logging.getLogger(__name__)

class Command(Enum):
    """Command."""

    W = 0x01    # Write
    RB = 0x02   # Read Buffer
    NOP = 0x03
    EW = 0x05   # Erase / Write
    RM = 0x06   # Read Modified
    EWA = 0x0d  # Erase / Write Alternate
    RMA = 0x0e  # Read Modified All
    EAU = 0x0f  # Erase All Unprotected
    WSF = 0x11  # Write Structured Field

COMMAND_MAP = {
    **{command.value: command for command in Command},

    # SNA
    0x63: Command.RMA,
    0x6f: Command.EAU,
    0x7e: Command.EWA,
    0xf1: Command.W,
    0xf2: Command.RB,
    0xf3: Command.WSF,
    0xf5: Command.EW,
    0xf6: Command.RM
}

class WCC:
    """Write control character."""

    def __init__(self, value):
        # TODO: Validate input.
        self.value = value

        self.reset = bool(value & 0x40)
        self.alarm = bool(value & 0x04)
        self.unlock_keyboard = bool(value & 0x02)
        self.reset_modified = bool(value & 0x01)

    def __repr__(self):
        return (f'<WCC reset={self.reset}, alarm={self.alarm}, '
                f'unlock_keyboard={self.unlock_keyboard}, '
                f'reset_modified={self.reset_modified}>')

class Order(Enum):
    """Order."""

    PT = 0x05   # Program Tab
    GE = 0x08   # Graphic Escape
    SBA = 0x11  # Set Buffer Address
    EUA = 0x12  # Erase Unprotected to Address
    IC = 0x13   # Insert Cursor
    SF = 0x1d   # Start Field
    SA = 0x28   # Set Attribute
    SFE = 0x29  # Start Field Extended
    MF = 0x2c   # Modify Field
    RA = 0x3c   # Repeat to Address

ORDERS = {order.value for order in Order}

class AID(Enum):
    """Attention identifier."""

    NONE = 0x60
    STRUCTURED_FIELD = 0x88
    CLEAR = 0x6d
    ENTER = 0x7d
    PA1 = 0x6c
    PA2 = 0x6e
    PA3 = 0x6b
    PF1 = 0xf1
    PF2 = 0xf2
    PF3 = 0xf3
    PF4 = 0xf4
    PF5 = 0xf5
    PF6 = 0xf6
    PF7 = 0xf7
    PF8 = 0xf8
    PF9 = 0xf9
    PF10 = 0x7a
    PF11 = 0x7b
    PF12 = 0x7c
    PF13 = 0xc1
    PF14 = 0xc2
    PF15 = 0xc3
    PF16 = 0xc4
    PF17 = 0xc5
    PF18 = 0xc6
    PF19 = 0xc7
    PF20 = 0xc8
    PF21 = 0xc9
    PF22 = 0x4a
    PF23 = 0x4b
    PF24 = 0x4c

SHORT_READ_AIDS = [AID.CLEAR, AID.PA1, AID.PA2, AID.PA3]

def parse_outbound_message(bytes_):
    """Parse a message from the host."""
    command_byte = bytes_[0]

    command = COMMAND_MAP.get(command_byte)

    if command is None:
        raise ValueError(f'Unrecognized command 0x{command_byte:02x}')

    if command in [Command.W, Command.EW, Command.EWA]:
        # TODO: Validate size.

        wcc = WCC(bytes_[1])
        orders = list(parse_orders(bytes_[2:]))

        return (command, wcc, orders)

    if command == Command.WSF:
        structured_fields = list(parse_outbound_structured_fields(bytes_[1:]))

        return (command, structured_fields)

    return (command,)

def format_inbound_read_buffer_message(aid, cursor_address, orders):
    """Format a read buffer message for the host."""
    bytes_ = bytearray()

    for (order, data) in orders:
        if order == Order.SF:
            bytes_.extend([Order.SF.value, data[0].value])
        elif order == Order.GE:
            bytes_.extend([Order.GE.value, data[0]])
        elif order is None:
            bytes_ += data
        else:
            raise NotImplementedError(f'{order} is not supported')

    return _format_inbound_message(aid, cursor_address, bytes_)

def format_inbound_read_modified_message(aid, cursor_address, orders, all_=False):
    """Format a read modified message for the host."""
    if aid in SHORT_READ_AIDS and not all_:
        return bytearray([aid.value])

    bytes_ = bytearray()

    for (order, data) in orders:
        if order == Order.SBA:
            bytes_.append(Order.SBA.value)
            bytes_.extend(format_address(data[0]))
        elif order == Order.SF:
            bytes_.extend([Order.SF.value, data[0].value])
        elif order == Order.GE:
            bytes_.extend([Order.GE.value, data[0]])
        elif order is None:
            bytes_ += data
        else:
            raise NotImplementedError(f'{order} is not supported')

    return _format_inbound_message(aid, cursor_address, bytes_)

def parse_orders(bytes_):
    """Parse orders from the host."""
    data = bytearray()

    index = 0

    while index < len(bytes_):
        byte = bytes_[index]

        if byte in ORDERS:
            if data:
                yield (None, data)

                data = bytearray()

            order = Order(byte)
            parameters = None

            index += 1

            if order == Order.PT:
                pass
            elif order == Order.GE:
                # TODO: validate size
                parameters = [bytes_[index]]
                index += 1
            elif order == Order.SBA:
                # TODO: validate size
                parameters = [parse_address(bytes_[index:index+2])[0]]
                index += 2
            elif order == Order.EUA:
                # TODO: validate size
                parameters = [parse_address(bytes_[index:index+2])[0]]
                index += 2
            elif order == Order.IC:
                pass
            elif order == Order.SF:
                parameters = [Attribute(bytes_[index])]
                index += 1
            elif order == Order.SA:
                # TODO: validate size
                parameters = [parse_extended_attribute(bytes_[index:index+2])]
                index += 2
            elif order in [Order.SFE, Order.MF]:
                # TODO: validate size
                attribute = None
                extended_attributes = []

                count = bytes_[index]
                index += 1

                for attribute_index in range(index, index + (count * 2), 2):
                    if bytes_[attribute_index] == 0xc0:
                        attribute = Attribute(bytes_[attribute_index+1])
                    else:
                        extended_attributes.append(parse_extended_attribute(bytes_[attribute_index:attribute_index+2]))

                parameters = [attribute, extended_attributes]
                index += count * 2
            elif order == Order.RA:
                # TODO: validate size
                stop_address = parse_address(bytes_[index:index+2])[0]
                index += 2

                # Peek ahead to detect a GE order.
                is_ge = False

                if bytes_[index] == Order.GE.value:
                    is_ge = True
                    index += 1

                parameters = [stop_address, bytes_[index], is_ge]
                index += 1

            yield (order, parameters)
        else:
            if byte == 0x00 or (byte >= 0x40 and byte <= 0xfe):
                data.append(byte)
            else:
                logger.warning(f'Value 0x{byte:02x} out of range')

            index += 1

    if data:
        yield (None, data)

def parse_outbound_structured_fields(bytes_):
    """Parse structured fields from the host."""
    index = 0

    while index < len(bytes_):
        remaining_length = len(bytes_) - index

        if remaining_length < 2:
            raise Exception('Invalid structured field')

        length = (bytes_[index] << 8) | bytes_[index+1]

        if length == 0:
            length = remaining_length

        if length < 3:
            raise Exception(f'Invalid structured field length: {length} must be at least 3')

        if length > remaining_length:
            raise Exception(f'Invalid structured field length: {length} greater than remaining {remaining_length} bytes')

        id_ = bytes_[index+2]

        data_length = length - 3
        data = bytes_[index+3:index+3+data_length]

        yield (id_, data)

        index += length

def format_inbound_structured_fields(structured_fields):
    """Format structured fields for the host."""
    bytes_ = bytearray([AID.STRUCTURED_FIELD.value])

    for (id_, data) in structured_fields:
        length = len(data) + 3

        bytes_.extend([(length >> 8) & 0xff, length & 0xff, id_])

        bytes_ += data

    return bytes_

def parse_extended_attribute(bytes_):
    """Parse an extended attribute."""

    if len(bytes_) != 2:
        raise Exception('Invalid extended attribute')

    type_ = bytes_[0]
    value = bytes_[1]

    if type_ == ExtendedAttributeType.ALL:
        # TODO: value should be 0x00
        # TODO: This should only be valid for SA orders...
        return AllExtendedAttribute(type_, value)
    elif type_ == ExtendedAttributeType.HIGHLIGHT:
        return HighlightExtendedAttribute(type_, value)
    elif type_ == ExtendedAttributeType.FOREGROUND_COLOR:
        return ForegroundColorExtendedAttribute(type_, value)

    logger.warning(f'Extended attribute 0x{type_:02x} not supported')

    return ExtendedAttribute(type_, value)

def parse_address(bytes_, size=None):
    """Parse an address."""
    if size == 16:
        return((bytes_[0] << 8) | bytes_[1], 16)

    setting = (bytes_[0] & 0xc0) >> 6

    # Handle a 12-bit address.
    if setting in [0b01, 0b11]:
        return (((bytes_[0] & 0x3f) << 6) | (bytes_[1] & 0x3f), 12)

    # Assume a 14-bit address.
    return (((bytes_[0] & 0x3f) << 8) | bytes_[1], 14)

# https://www.tommysprinkle.com/mvs/P3270/iocodes.htm
SIX_BIT_CHARACTER_MAP = [
    0x40, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7,
    0xc8, 0xc9, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f,
    0x50, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7,
    0xd8, 0xd9, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f,
    0x60, 0x61, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7,
    0xe8, 0xe9, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f,
    0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7,
    0xf8, 0xf9, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f
]

def format_address(address, size=12):
    """Format an address."""

    # TODO: Validate that the address is within range based on size.

    if size == 16:
        return bytes([(address >> 8) & 0xff, address & 0xff])

    if size == 14:
        return bytes([(address >> 8) & 0x3f, address & 0xff])

    if size == 12:
        return bytes([SIX_BIT_CHARACTER_MAP[(address >> 6) & 0x3f],
                      SIX_BIT_CHARACTER_MAP[address & 0x3f]])

    raise ValueError('Invalid size')

def _format_inbound_message(aid, cursor_address, data_bytes):
    message_bytes = bytearray([aid.value])

    message_bytes += format_address(cursor_address)
    message_bytes += data_bytes

    return message_bytes



"""
tn3270.telnet
~~~~~~~~~~~~~
"""

import time
import logging
import socket
import selectors
from telnetlib import IAC, WILL, WONT, DO, DONT, SB, SE, BINARY, EOR, TTYPE, TN3270E

# https://tools.ietf.org/html/rfc855
RFC855_EOR = b'\xef'

# https://tools.ietf.org/html/rfc1091
RFC1091_IS = b'\x00'
RFC1091_SEND = b'\x01'

# https://tools.ietf.org/html/rfc2355
RFC2355_CONNECT = b'\x01'
RFC2355_DEVICE_TYPE = b'\x02'
RFC2355_FUNCTIONS = b'\x03'
RFC2355_IS = b'\x04'
RFC2355_REJECT = b'\x06'
RFC2355_REQUEST = b'\x07'
RFC2355_SEND = b'\x08'

class Telnet:
    """TN3270 client."""

    def __init__(self, terminal_type, is_tn3270e_enabled=True):
        self.logger = logging.getLogger(__name__)

        self.terminal_type = terminal_type
        self.is_tn3270e_enabled = is_tn3270e_enabled

        self.socket = None
        self.socket_selector = None
        self.eof = None

        self.host_options = set()
        self.client_options = set()
        self.is_tn3270e_negotiated = False
        self.device_type = None
        self.device_name = None

        self.buffer = bytearray()
        self.iac_buffer = bytearray()
        self.records = []

    def open(self, host, port, tn3270_negotiation_timeout=None, ssl_context=None, ssl_server_hostname=None):
        """Open the connection."""
        self.close()

        self.socket = socket.create_connection((host, port))

        if ssl_context:
            self.socket = ssl_context.wrap_socket(self.socket, server_hostname=ssl_server_hostname)

        self.socket_selector = selectors.DefaultSelector()

        self.socket_selector.register(self.socket, selectors.EVENT_READ)

        self.eof = False

        self.host_options = set()
        self.client_options = set()
        self.is_tn3270e_negotiated = False
        self.device_type = None
        self.device_name = None

        self.buffer = bytearray()
        self.iac_buffer = bytearray()
        self.records = []

        self._negotiate_tn3270(timeout=tn3270_negotiation_timeout)

    def close(self):
        """Close the connection."""
        if self.socket_selector is not None:
            self.socket_selector.unregister(self.socket)

        if self.socket is not None:
            self.socket.close()

            self.socket = None

        if self.socket_selector is not None:
            self.socket_selector.close()

            self.socket_selector = None

    def read_multiple(self, limit=None, timeout=None):
        """Read multiple records."""
        records = self._read_multiple_buffered(limit)

        if records:
            return records

        self._read_while(lambda: not self.eof and not self.records, timeout)

        # TODO: Determine what happens to any bytes in the buffer if EOF is
        # encountered without EOR - should that yield a record?
        if self.eof and self.buffer:
            self.logger.warning('EOF encountered with partial record')

        return self._read_multiple_buffered(limit)

    def write(self, record):
        """Write a record."""

        # Add 3270-DATA TN3270E header if in TN3270E mode.
        if self.is_tn3270e_negotiated:
            record = bytes([0x00, 0x00, 0x00, 0x00, 0x00]) + record

        self.socket.sendall(record.replace(IAC, IAC * 2) + IAC + RFC855_EOR)

    @property
    def is_tn3270_negotiated(self):
        """Has TN3270 or TN3270E mode been negotiated."""

        if self.is_tn3270e_negotiated:
            return True

        # https://tools.ietf.org/html/rfc1576
        return (self.client_options.issuperset([BINARY, EOR, TTYPE])
                and self.host_options.issuperset([BINARY, EOR]))

    def _read(self, timeout):
        if self.eof:
            raise EOFError

        if not self.socket_selector.select(timeout):
            return

        bytes_ = self.socket.recv(1024)

        if not bytes_:
            self.eof = True
            return

        for byte in bytes_:
            self._feed(bytes([byte]))

    def _read_while(self, predicate, timeout):
        remaining_timeout = timeout

        while predicate():
            read_time = time.perf_counter()

            self._read(remaining_timeout)

            if remaining_timeout is not None:
                remaining_timeout -= (time.perf_counter() - read_time)

                if remaining_timeout < 0:
                    break

    def _read_multiple_buffered(self, limit=None):
        if self.eof and not self.records:
            raise EOFError

        if not self.records:
            return []

        count = limit if limit is not None else len(self.records)

        records = self.records[:count]

        self.records = self.records[count:]

        return records

    def _feed(self, byte):
        if not self.iac_buffer:
            if byte == IAC:
                self.iac_buffer += byte
                return

            self.buffer += byte
        elif len(self.iac_buffer) == 1:
            if byte == IAC:
                self.buffer += IAC
                self.iac_buffer.clear()
                return

            if byte == RFC855_EOR:
                self._eor(bytearray(self.buffer))

                self.buffer.clear()
                self.iac_buffer.clear()
                return

            if byte in [WILL, WONT, DO, DONT, SB]:
                self.iac_buffer += byte
                return

            self.logger.warning(f'Unexpected byte 0x{byte[0]:02x} in IAC state')

            self.iac_buffer.clear()
        elif len(self.iac_buffer) > 1:
            command = self.iac_buffer[1:2]

            if command in [WILL, WONT, DO, DONT]:
                self._handle_negotiation(command, byte)

                self.iac_buffer.clear()
                return

            if command == SB:
                if byte == SE:
                    if self.iac_buffer[-1:] != IAC:
                        self.logger.warning('Expected IAC prior to SE')

                    self._handle_subnegotiation(self.iac_buffer[2:-1].replace(IAC * 2, IAC))

                    self.iac_buffer.clear()
                    return

                self.iac_buffer += byte
                return

            self.logger.warning(f'Unrecognized command 0x{command:02x}')

            self.iac_buffer.clear()

    def _handle_negotiation(self, command, option):
        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug((f'Negotiate: Command = 0x{command.hex()}, '
                               f'Option = 0x{option.hex()}'))

        if command == WILL:
            if option in [BINARY, EOR, TTYPE] or (option == TN3270E and self.is_tn3270e_enabled):
                self.host_options.add(option)

                self.socket.sendall(IAC + DO + option)
            else:
                self.socket.sendall(IAC + DONT + option)
        elif command == WONT:
            self.host_options.discard(option)

            self.socket.sendall(IAC + DONT + option)
        elif command == DO:
            if option in [BINARY, EOR, TTYPE] or (option == TN3270E and self.is_tn3270e_enabled):
                self.client_options.add(option)

                self.socket.sendall(IAC + WILL + option)
            else:
                self.socket.sendall(IAC + WONT + option)
        elif command == DONT:
            self.client_options.discard(option)

            self.socket.sendall(IAC + WONT + option)

    def _handle_subnegotiation(self, bytes_):
        if bytes_ == TTYPE + RFC1091_SEND:
            self.logger.debug('Received TTYPE SEND request')

            terminal_type = self.terminal_type.encode('ascii')

            self.socket.sendall(IAC + SB + TTYPE + RFC1091_IS + terminal_type + IAC + SE)
        elif bytes_.startswith(TN3270E):
            if not self.is_tn3270e_enabled:
                self.logger.warning('TN3270E subnegotiation requested but TN3270E not enabled')

            self._handle_tn3270e_subnegotiation(bytes_[1:])

    def _handle_tn3270e_subnegotiation(self, bytes_):
        if bytes_ == RFC2355_SEND + RFC2355_DEVICE_TYPE:
            self.logger.debug('Received TN3270E SEND DEVICE-TYPE request')

            device_type = self.terminal_type.replace('IBM-3279', 'IBM-3278').encode('ascii')

            self.socket.sendall(IAC + SB + TN3270E + RFC2355_DEVICE_TYPE + RFC2355_REQUEST + device_type + IAC + SE)
        elif bytes_.startswith(RFC2355_DEVICE_TYPE + RFC2355_IS):
            self.logger.debug('Received TN3270E DEVICE-TYPE response')

            (device_type, device_name) = bytes_[2:].split(RFC2355_CONNECT)

            self.device_type = device_type.decode('ascii')
            self.device_name = device_name.decode('ascii')

            # Request basic TN3270E, no functions...
            self.socket.sendall(IAC + SB + TN3270E + RFC2355_FUNCTIONS + RFC2355_REQUEST + IAC + SE)

        elif bytes_.startswith(RFC2355_DEVICE_TYPE + RFC2355_REJECT):
            self.logger.debug('Received TN3270E DEVICE-TYPE REJECT response')

            # TODO: Implement further DEVICE-TYPE negotiation...
            self.logger.warning('TN3270E DEVICE-TYPE request rejected, complete negotiation not implemented')

            self.socket.sendall(IAC + WONT + TN3270E)
        elif bytes_.startswith(RFC2355_FUNCTIONS + RFC2355_REQUEST):
            self.logger.debug('Received TN3270E FUNCTIONS request')

            functions = set(bytes_[2:])

            if functions:
                self.socket.sendall(IAC + SB + TN3270E + RFC2355_FUNCTIONS + RFC2355_REQUEST + IAC + SE)
            else:
                self.logger.debug('TN3270E negotiation complete')

                self.is_tn3270e_negotiated = True
        elif bytes_.startswith(RFC2355_FUNCTIONS + RFC2355_IS):
            self.logger.debug('Received TN3270E FUNCTIONS response')

            functions = set(bytes_[2:])

            if functions:
                self.logger.warning('TN3270E FUNCTIONS response contains unrequested functions, aborting TN3270E negotiation')

                self.socket.sendall(IAC + WONT + TN3270E)
            else:
                self.logger.debug('TN3270E negotiation complete')

                self.is_tn3270e_negotiated = True

    def _negotiate_tn3270(self, timeout):
        self._read_while(lambda: not self.is_tn3270_negotiated and not self.eof
                         and not self.buffer, timeout)

        if not self.is_tn3270_negotiated:
            raise Exception('Unable to negotiate TN3270 mode')

    def _eor(self, record):
        self.logger.debug('Received EOR')

        if self.is_tn3270e_negotiated:
            data_type = record[0]

            if data_type == 0x00:
                self.records.append(record[5:])
            else:
                self.logger.warning(f'Unsupported TN3270E DATA-TYPE 0x{data_type:02x}')
        else:
            self.records.append(record)

    def __del__(self):
        self.close()



import os
from setuptools import setup

ABOUT = {}

with open(os.path.join(os.path.dirname(__file__), "tn3270", "__about__.py")) as file:
    exec(file.read(), ABOUT)

LONG_DESCRIPTION = """# pytn3270

Python TN3270 library.

See [GitHub](https://github.com/lowobservable/pytn3270#readme) for more information.
"""

setup(
    name='pytn3270',
    version=ABOUT['__version__'],
    description='TN3270 library',
    url='https://github.com/lowobservable/pytn3270',
    author='Andrew Kay',
    author_email='projects@ajk.me',
    packages=['tn3270'],
    install_requires=[],
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Communications',
        'Topic :: Terminals'
    ]
)
